import numpy as np
from .math import sigmoid

def parse_gaussian_splatting_data(ply_data):
    vertices = ply_data['vertex']
    
    positions = np.vstack([vertices['x'], vertices['y'], vertices['z']]).T  # (N, 3)
    opacities = vertices['opacity']  # (N,)
    
    # Extracting Gaussian parameters
    # Note: Different implementations may use different naming conventions
    scales = np.vstack([vertices['scale_0'], vertices['scale_1'], vertices['scale_2']]).T  # Scaling (N, 3)
    rotations = np.vstack([vertices['rot_0'], vertices['rot_1'], vertices['rot_2'], vertices['rot_3']]).T  # Rotation quaternions (N, 4)
    
    # Extracting Spherical Harmonics (SH) coefficients:
    sh_coeffs = []
    i = 0
    while True:
        if i < 3:  # First 3 coefficients are DC terms (base color)
            key = f'f_dc_{i}'
        else:  # Remaining coefficients represent higher-frequency components
            key = f'f_rest_{i-3}'
        if not key in vertices:
            break
        sh_coeffs.append(vertices[key])
        i += 1
    
    sh_coeffs =np.array(sh_coeffs).T
    sh_coeffs = np.reshape(sh_coeffs, (sh_coeffs.shape[0], -1, 3))  # RGB triplets (3 channels) (N, -, 3)

    scales = np.exp(scales)
    sh_coeffs = sigmoid(sh_coeffs)
    opacities = sigmoid(opacities)
    
    return {
        'positions': positions,
        'opacities': opacities,
        'scales': scales,
        'rotations': rotations,
        'sh_coeffs': sh_coeffs
    }



import numpy as np
import taichi as ti

__version__ = "1.6.0"

class Scene():
    def __init__(
        self,
        smoke_density_taichi,
        smoke_color_taichi,
        point_lights_pos_taichi,
        point_lights_intensity_taichi,
        lighting=True,
        index_of_refraction_taichi=None,
        ray_tracing_stop_threshold=0.01,  # 0 ~ 1
        background=[0.2, 0.2, 0.2],
        smoke_density_factor=1.,
        ray_tracing_step_size_factor=1.,  # The smaller the value here, the higher the ray tracing quality.
        light_ray_tracing_step_size_factor=3.,  # The smaller the value here, the higher the shadow quality.
        ray_tracing_max_steps=10000  # This only takes effect in scenes where light rays may bend, such as those containing refractive materials.
    ):
        # Volume data
        self.smoke_density = smoke_density_taichi  # Smoke density
        self._smoke_density_factor = ti.field(dtype=ti.f32, shape=())
        self._smoke_density_factor[None] = smoke_density_factor
        self.smoke_color = smoke_color_taichi  # Smoke color
        self.index_of_refraction = index_of_refraction_taichi

        # Light
        self.point_lights_pos = point_lights_pos_taichi
        self.point_lights_intensity = point_lights_intensity_taichi
    
        # Camera
        self._fov = ti.field(dtype=ti.f32, shape=())
        self._fov[None] = 0.5924  # 2 * tan(vertical field of view / 2)
        self._camera_distance = ti.field(dtype=ti.f32, shape=())
        self._camera_distance[None] = 3
        self._camera_phi = ti.field(dtype=ti.f32, shape=())
        self._camera_theta = ti.field(dtype=ti.f32, shape=())
        self._background = ti.Vector.field(3, dtype=ti.f32, shape=())
        self._background[None] = ti.Vector(background)

        # Ray tracing
        self._step_length = ti.field(dtype=ti.f32, shape=())
        pixel_size = 1 / np.max(smoke_density_taichi.shape)
        self._step_length[None] = pixel_size * ray_tracing_step_size_factor
        self._step_length_light = ti.field(dtype=ti.f32, shape=())
        self._step_length_light[None] = pixel_size * light_ray_tracing_step_size_factor
        self._stop_threshold = ti.field(dtype=ti.f32, shape=())
        self._stop_threshold[None] = ray_tracing_stop_threshold  # Terminate ray tracing when the accumulated transparency of the view ray falls below this value.

        # Light density in volume
        self.light_density = ti.Vector.field(3, dtype=ti.f32, shape=smoke_density_taichi.shape)
        self.light_density.from_numpy(np.ones(list(smoke_density_taichi.shape) + [3]))

        if lighting:
            @ti.kernel
            def update_light():  # Update shadow.
                for i, j, k in self.light_density:
                    self.light_density[i, j, k] = ti.Vector([0., 0., 0.])
                    pos = ti.Vector([i + 0.5, j + 0.5, k + 0.5]) / self.smoke_density.shape - 0.5
                    for l in ti.ndrange(self.point_lights_pos.shape[0]):
                        d = self.point_lights_pos[l] - pos
                        distance_squared = ti.math.dot(d, d)
                        transmittance = 1.
                        d = d.normalized()
                        pos_2 = pos
                        # pos_2 += d * (pixel_size * 0.5)
                        while True:
                            if pos_2.x > 0.5 and d.x > 0 or pos_2.x < -0.5 and d.x < 0:
                                break
                            if pos_2.y > 0.5 and d.y > 0 or pos_2.y < -0.5 and d.y < 0:
                                break
                            if pos_2.z > 0.5 and d.z > 0 or pos_2.z < -0.5 and d.z < 0:
                                break
                            pos_maped = (pos_2 + 0.5) * self.smoke_density.shape
                            x_int = int(pos_maped.x)
                            y_int = int(pos_maped.y)
                            z_int = int(pos_maped.z)
                            if x_int >= 0 and x_int < self.smoke_density.shape[0] and y_int >= 0 and y_int < self.smoke_density.shape[1] and z_int >= 0 and z_int < self.smoke_density.shape[2]:
                                transmittance *= 1 - self._smoke_density_factor[None] * self.smoke_density[x_int, y_int, z_int] * self._step_length_light[None]
                            pos_2 += d * self._step_length_light[None]
                        self.light_density[i, j, k] += self.point_lights_intensity[l] * (transmittance / distance_squared)
            self.update_light = update_light
        else:
            @ti.kernel
            def update_light():  # Update shadow.
                pass
            self.update_light = update_light

        if self.index_of_refraction is None:
            @ti.func
            def ray_tracing_one_step(pos, d, pixels_color, transmittance):
                pos_maped = (pos + 0.5) * self.smoke_density.shape
                x_int = int(pos_maped.x)
                y_int = int(pos_maped.y)
                z_int = int(pos_maped.z)
                if x_int >= 0 and x_int < self.smoke_density.shape[0] and y_int >= 0 and y_int < self.smoke_density.shape[1] and z_int >= 0 and z_int < self.smoke_density.shape[2]:
                    transmittance *= 1 - self._smoke_density_factor[None] * self.smoke_density[x_int, y_int, z_int] * self._step_length[None]
                    pixels_color += self._smoke_density_factor[None] * self.smoke_density[x_int, y_int, z_int] * self.smoke_color[x_int, y_int, z_int] * self._step_length[None] * self.light_density[x_int, y_int, z_int] * transmittance
                pos += d * self._step_length[None]
                return pos, d, pixels_color, transmittance, False
            self.ray_tracing_one_step = ray_tracing_one_step
        else:
            @ti.func
            def ray_tracing_one_step(pos, d, pixels_color, transmittance):
                pos_maped = (pos + 0.5) * self.smoke_density.shape
                x_int = int(pos_maped.x)
                y_int = int(pos_maped.y)
                z_int = int(pos_maped.z)
                to_break = False
                if x_int >= 0 and x_int < self.smoke_density.shape[0] and y_int >= 0 and y_int < self.smoke_density.shape[1] and z_int >= 0 and z_int < self.smoke_density.shape[2]:
                    transmittance *= 1 - self._smoke_density_factor[None] * self.smoke_density[x_int, y_int, z_int] * self._step_length[None]
                    pixels_color += self._smoke_density_factor[None] * self.smoke_density[x_int, y_int, z_int] * self.smoke_color[x_int, y_int, z_int] * self._step_length[None] * self.light_density[x_int, y_int, z_int] * transmittance

                    if x_int >= 1 and x_int < self.smoke_density.shape[0] - 1 and y_int >= 1 and y_int < self.smoke_density.shape[1] - 1 and z_int >= 1 and z_int < self.smoke_density.shape[2] - 1:
                        index_of_refraction = self.index_of_refraction[x_int, y_int, z_int]
                        index_of_refraction_grad = ti.Vector([
                            (self.index_of_refraction[x_int + 1, y_int, z_int] - self.index_of_refraction[x_int - 1, y_int, z_int]),
                            (self.index_of_refraction[x_int, y_int + 1, z_int] - self.index_of_refraction[x_int, y_int - 1, z_int]),
                            (self.index_of_refraction[x_int, y_int, z_int + 1] - self.index_of_refraction[x_int, y_int, z_int - 1])
                        ]) / (pixel_size * 2.)
                        index_of_refraction_change = ti.math.dot(d, index_of_refraction_grad) * self._step_length[None]
                        if index_of_refraction_change != 0.:
                            normal = index_of_refraction_grad.normalized()
                            relative_index_of_refraction = (index_of_refraction + index_of_refraction_change * 0.5) / (index_of_refraction - index_of_refraction_change * 0.5)
                            d_n = ti.math.dot(d, normal) * normal
                            d_t = d - d_n
                            cos_old = ti.abs(ti.math.dot(d, normal))
                            sin_old = (1 - cos_old ** 2) ** 0.5
                            sin_new = sin_old / relative_index_of_refraction
                            if sin_new > 1.:
                                to_break = True
                            else:
                                cos_new = (1 - sin_new ** 2) ** 0.5
                                d_n *= cos_new / cos_old
                                d_t *= sin_new / sin_old
                                d = d_n + d_t
                                d = d.normalized()

                pos += d * self._step_length[None]
                return pos, d, pixels_color, transmittance, to_break
            self.ray_tracing_one_step = ray_tracing_one_step

        @ti.func
        def ray_tracing(pos, d):
            pixels_color = ti.Vector([0., 0., 0.])
            transmittance = 1.
            distance_to_sphere = self._camera_distance[None] - 0.866025  # The constant here is 0.5 * 2 ** 0.5
            if distance_to_sphere > 0:
                pos += d * distance_to_sphere
            i = ray_tracing_max_steps
            while i > 0:
                if (pos.x > 0.5 and d.x > 0 or pos.x < -0.5 and d.x < 0) or (pos.y > 0.5 and d.y > 0 or pos.y < -0.5 and d.y < 0) or (pos.z > 0.5 and d.z > 0 or pos.z < -0.5 and d.z < 0):
                    break
                if transmittance < self._stop_threshold[None]:
                    break

                pos, d, pixels_color, transmittance, to_break = self.ray_tracing_one_step(pos, d, pixels_color, transmittance)
                if to_break:
                    break

                i -= 1
            
            pixels_color += self._background[None] * transmittance
            return pixels_color
        self.ray_tracing = ray_tracing

        @ti.kernel
        def render(pixels: ti.template()):  # type: ignore
            camera_pos = self._camera_distance[None] * ti.Vector([
                ti.cos(self._camera_phi[None]) * ti.cos(self._camera_theta[None]),
                ti.sin(self._camera_phi[None]) * ti.cos(self._camera_theta[None]),
                ti.sin(self._camera_theta[None])
            ])
            camera_u_vector = ti.Vector([
                -ti.sin(self._camera_phi[None]),
                ti.cos(self._camera_phi[None]),
                0
            ])
            camera_v_vector = ti.Vector([
                -ti.cos(self._camera_phi[None]) * ti.sin(self._camera_theta[None]),
                -ti.sin(self._camera_phi[None]) * ti.sin(self._camera_theta[None]),
                ti.cos(self._camera_theta[None])
            ])
            camera_direction = -camera_pos / self._camera_distance[None]

            for i, j in pixels:
                pos = camera_pos
                d = camera_direction + camera_u_vector * (self._fov[None] * (i - pixels.shape[0] / 2) / pixels.shape[1]) + camera_v_vector * (self._fov[None] * (j / pixels.shape[1] - 0.5))
                d = d.normalized()
                
                pixels[i, j] = self.ray_tracing(pos, d)
        self.render = render
    
    @property
    def smoke_density_factor(self):
        return self._smoke_density_factor[None]

    @smoke_density_factor.setter
    def smoke_density_factor(self, value):
        self._smoke_density_factor[None] = value

    def get_vertical_field_of_view(self, degrees=True):  # Get vertical field of view. Default is 33°.
        return np.atan(self._fov[None] / 2) * 2 * (180 / np.pi if degrees else 1)

    def set_vertical_field_of_view(self, angle, degrees=True):  # Set vertical field of view. Default is 33°.
        self._fov[None] = 2 * np.tan(angle * (np.pi / 180 if degrees else 1) / 2)

    def get_camera_phi(self, degrees=True):
        return self._camera_phi[None] * (180 / np.pi if degrees else 1)

    def set_camera_phi(self, angle, degrees=True):
        self._camera_phi[None] = angle * (np.pi / 180 if degrees else 1)

    def get_camera_theta(self, degrees=True):
        return self._camera_theta[None] * (180 / np.pi if degrees else 1)

    def set_camera_theta(self, angle, degrees=True):
        self._camera_theta[None] = angle * (np.pi / 180 if degrees else 1)
        if self._camera_theta[None] < np.pi * -0.5:
            self._camera_theta[None] = np.pi * -0.5
        if self._camera_theta[None] > np.pi * 0.5:
            self._camera_theta[None] = np.pi * 0.5
    
    @property
    def camera_distance(self):
        return self._camera_distance[None]

    @camera_distance.setter
    def camera_distance(self, value):
        self._camera_distance[None] = value
    
    @property
    def background(self):
        return list(self._background[None])

    @background.setter
    def background(self, value):
        self._background[None] = ti.Vector(value)
    
    @property
    def step_length(self):
        return self._step_length[None]

    @step_length.setter
    def step_length(self, value):
        self._step_length[None] = value
    
    @property
    def step_length_light(self):
        return self._step_length_light[None]

    @step_length_light.setter
    def step_length_light(self, value):
        self._step_length_light[None] = value
    
    @property
    def stop_threshold(self):
        return self._stop_threshold[None]

    @stop_threshold.setter
    def stop_threshold(self, value):
        self._stop_threshold[None] = value

class DisplayWindow():
    def __init__(
        self,
        smoke_density,  # Can be NumPy array or Taichi field.
        smoke_color=None,  # Can be None, NumPy array or Taichi vector field. If left None, uniform white applied.
        index_of_refraction=None,  # Can be None, NumPy array or Taichi vector field.
        point_lights_pos=None,  # Can be None, NumPy array or Taichi vector field. If left None, default lights applied.
        point_lights_intensity=None,  # Can be None, NumPy array or Taichi vector field. If left None, default lights applied.
        lighting=True,
        resolution=(720, 720),
        ray_tracing_stop_threshold=0.01,  # 0 ~ 1
        background=[0.2, 0.2, 0.2],
        init_taichi=True,
        taichi_arch=ti.gpu,
        smoke_density_factor=1.,
        ray_tracing_step_size_factor=1.,  # The smaller the value here, the higher the ray tracing quality.
        light_ray_tracing_step_size_factor=3.,  # The smaller the value here, the higher the shadow quality.
        ray_tracing_max_steps=10000  # This only takes effect in scenes where light rays may bend, such as those containing refractive materials.
    ):
        if init_taichi:
            ti.init(arch=taichi_arch)
        
        if not isinstance(smoke_density, ti.Field):
            smoke_density_numpy = smoke_density
            smoke_density = ti.field(dtype=ti.f32, shape=smoke_density_numpy.shape)
            smoke_density.from_numpy(smoke_density_numpy)

        if smoke_color is None:
            smoke_color = np.ones(list(smoke_density.shape) + [3])
        if not isinstance(smoke_color, ti.Field):
            smoke_color_numpy = smoke_color
            smoke_color = ti.Vector.field(3, dtype=ti.f32, shape=smoke_color_numpy.shape[:-1])
            smoke_color.from_numpy(smoke_color_numpy)

        if not index_of_refraction is None:
            if not isinstance(index_of_refraction, ti.Field):
                index_of_refraction_numpy = index_of_refraction
                index_of_refraction = ti.field(dtype=ti.f32, shape=index_of_refraction_numpy.shape)
                index_of_refraction.from_numpy(index_of_refraction_numpy)

        if point_lights_pos is None:
            point_lights_pos = np.array([[0, 0, 5]], dtype=float)
        if not isinstance(point_lights_pos, ti.Field):
            point_lights_pos_numpy = point_lights_pos
            point_lights_pos = ti.Vector.field(3, dtype=ti.f32, shape=point_lights_pos_numpy.shape[:-1])
            point_lights_pos.from_numpy(point_lights_pos_numpy)

        if point_lights_intensity is None:
            point_lights_intensity = np.array([[50, 50, 50]], dtype=float)
        if not isinstance(point_lights_intensity, ti.Field):
            point_lights_intensity_numpy = point_lights_intensity
            point_lights_intensity = ti.Vector.field(3, dtype=ti.f32, shape=point_lights_intensity_numpy.shape[:-1])
            point_lights_intensity.from_numpy(point_lights_intensity_numpy)

        self.scene = Scene(
            smoke_density_taichi=smoke_density,
            smoke_color_taichi=smoke_color,
            index_of_refraction_taichi=index_of_refraction,
            point_lights_pos_taichi=point_lights_pos,
            point_lights_intensity_taichi=point_lights_intensity,
            lighting=lighting,
            ray_tracing_stop_threshold=ray_tracing_stop_threshold,
            background=background,
            smoke_density_factor=smoke_density_factor,
            ray_tracing_step_size_factor=ray_tracing_step_size_factor,
            light_ray_tracing_step_size_factor=light_ray_tracing_step_size_factor,
            ray_tracing_max_steps=ray_tracing_max_steps)

        # Window
        self.resolution = tuple(resolution)
        self.pixels = ti.Vector.field(3, dtype=ti.f32, shape=resolution)

        # Interaction
        self.mouse_pressed = False
        self.cursor_start_pos = (-1, -1)
        self.camera_rotation_speed = 230.  # Unit: degree pre image width or height
        self.camera_zooming_speed = 0.0007
    
    def mouse_pressed_event(self, pos):
        pass

    def mouse_drag_event(self, pos, pos_delta):
        self.scene.set_camera_phi(self.scene.get_camera_phi() - pos_delta[0] * self.camera_rotation_speed)
        self.scene.set_camera_theta(self.scene.get_camera_theta() - pos_delta[1] * self.camera_rotation_speed)

    def mouse_wheel_event(self, delta):
        self.scene.camera_distance *= (1 - self.camera_zooming_speed) ** delta[1]
    
    def show(
            self,
            title="Render",
            update_light_each_step=False,
            callback=None,  # Users can update smoke density, rotate camera etc. each step by assigning this callback function.
            image_process=None,  # Users can edit the rendering result before it displayed in the window each step by assigning this callback function.
            enable_mouse_rotating=True
        ):
        self.scene.update_light()  # Calculate light and shadow

        gui = ti.GUI(title, res=self.resolution)
        iteration = 0
        while gui.running:
            if update_light_each_step:
                self.scene.update_light()
            self.scene.render(self.pixels)
            if not image_process is None:
                image_process(iteration, self.pixels)
            gui.set_image(self.pixels)
            gui.show()

            # Manage mouse events
            if enable_mouse_rotating:
                if gui.is_pressed(ti.GUI.LMB):
                    cursor_pos = gui.get_cursor_pos()
                    if not self.mouse_pressed:
                        self.mouse_pressed_event(cursor_pos)
                        self.mouse_pressed = True
                        self.cursor_start_pos = cursor_pos
                    if self.cursor_start_pos[0] != cursor_pos[0] or self.cursor_start_pos[1] != cursor_pos[1]:
                        self.mouse_drag_event(cursor_pos, (cursor_pos[0] - self.cursor_start_pos[0], cursor_pos[1] - self.cursor_start_pos[1]))
                        self.cursor_start_pos = cursor_pos
                else:
                    self.mouse_pressed = False
                for event in gui.get_events(ti.GUI.WHEEL):
                    self.mouse_wheel_event(event.delta)

            if not callback is None:
                callback(iteration, self.scene)
            iteration += 1

def plot_volume(
    smoke_density=None,  # Can be NumPy array or Taichi field.
    smoke_color=None,  # Can be None, NumPy array or Taichi vector field. If left None, uniform white applied.
    index_of_refraction=None,  # Can be None, NumPy array or Taichi vector field.
    point_lights_pos=None,  # Can be None, NumPy array or Taichi vector field. If left None, default lights applied.
    point_lights_intensity=None,  # Can be None, NumPy array or Taichi vector field. If left None, default lights applied.
    lighting=True,
    resolution=(720, 720),
    ray_tracing_stop_threshold=0.01,  # 0 ~ 1
    background=[0.2, 0.2, 0.2],
    init_taichi=True,
    taichi_arch=ti.gpu,
    smoke_density_factor=1.,
    ray_tracing_step_size_factor=1.,  # The smaller the value here, the higher the ray tracing quality.
    light_ray_tracing_step_size_factor=3.,  # The smaller the value here, the higher the shadow quality.
    ray_tracing_max_steps=10000,  # This only takes effect in scenes where light rays may bend, such as those containing refractive materials.
    camera_phi=0,
    camera_theta=0,
    camera_distance=3,

    title="Render",
    update_light_each_step=False,
    callback=None,  # Users can update smoke density, rotate camera etc. each step by assigning this callback function.
    image_process=None,  # Users can edit the rendering result before it displayed in the window each step by assigning this callback function.
    enable_mouse_rotating=True
):
    window = DisplayWindow(
        smoke_density=smoke_density,
        smoke_color=smoke_color,
        index_of_refraction=index_of_refraction,
        point_lights_pos=point_lights_pos,
        point_lights_intensity=point_lights_intensity,
        lighting=lighting,
        resolution=resolution,
        ray_tracing_stop_threshold=ray_tracing_stop_threshold,
        background=background,
        init_taichi=init_taichi,
        taichi_arch=taichi_arch,
        smoke_density_factor=smoke_density_factor,
        ray_tracing_step_size_factor=ray_tracing_step_size_factor,
        light_ray_tracing_step_size_factor=light_ray_tracing_step_size_factor,
        ray_tracing_max_steps=ray_tracing_max_steps
    )
    window.scene.set_camera_phi(camera_phi)
    window.scene.set_camera_theta(camera_theta)
    window.scene.camera_distance = camera_distance

    window.show(
        title=title,
        update_light_each_step=update_light_each_step,
        callback=callback,
        image_process=image_process,
        enable_mouse_rotating=enable_mouse_rotating
    )



import numpy as np
import taichi as ti
from .math import compute_covariance_inv, rotation_matrix_to_quaternion, rotation_quaternion_to_matrix, quaternion_multiply

def construct_frame(x, y=None):
    x /= np.linalg.norm(x)
    for e in [y, [1, 0, 0], [0, 1, 0]]:
        if e is None:
            continue
        z = np.cross(x, e)
        z_norm = np.linalg.norm(z)
        if z_norm != 0:
            z /= z_norm
            y = np.cross(z, x)
            return x, y, z

def empty_canvas(resolution):
    if type(resolution) == int:
        smoke = ti.field(dtype=float, shape=[resolution, resolution, resolution])
        smoke_color = ti.Vector.field(3, dtype=float, shape=smoke.shape)
        smoke_color.from_numpy(np.ones(list(smoke_color.shape) + [3]))
        return smoke, smoke_color
    raise TypeError("Unsupported type of resolution: " + str(type(resolution)))@ti.kernel

@ti.kernel
def clean(
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template()  # type: ignore
    ):
    for I in ti.grouped(smoke_density_taichi):
        smoke_density_taichi[I] = 0
        smoke_color_taichi[I] = ti.Vector([1, 1, 1])

@ti.kernel
def multiply(
    taichi_field: ti.template(),  # type: ignore
    k: float
):
    for I in ti.grouped(taichi_field):
        taichi_field[I] *= k

@ti.kernel
def clip_kernel(
    taichi_field: ti.template(),  # type: ignore
    min: float,
    max: float
):
    for I in ti.grouped(taichi_field):
        taichi_field[I] = ti.math.clamp(taichi_field[I], min, max)

def clip(
    taichi_field,
    min=0,
    max=1
):
    clip_kernel(taichi_field, min, max)

@ti.kernel
def gamma(
    taichi_field: ti.template(),  # type: ignore
    power: float
):
    for I in ti.grouped(taichi_field):
        taichi_field[I] = taichi_field[I] ** power

@ti.func
def mix(color_1, density_1, color_2, density_2):
    return (color_1 * density_1 + color_2 * density_2) / (density_1 + density_2)

@ti.kernel
def fill_rectangle(  # Fill rectangle
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    start: ti.math.vec3,  # type: ignore
    scale: ti.math.vec3,  # type: ignore
    density: float,
    color: ti.math.vec3  # type: ignore
    ):
    end_int = int(ti.round(start + scale))
    start_int = int(ti.round(start))
    start_int = ti.math.max(0, start_int)
    end_int = ti.math.min(smoke_density_taichi.shape, end_int)
    for I in ti.grouped(ti.ndrange([start_int.x, end_int.x], [start_int.y, end_int.y], [start_int.z, end_int.z])):
        smoke_color_taichi[I] = mix(smoke_color_taichi[I], smoke_density_taichi[I], color, density)
        smoke_density_taichi[I] += density

@ti.kernel
def fill_disk(  # Fill disk (No anti-aliasing)
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    center: ti.math.vec3,  # type: ignore
    radius: float,
    density: float,
    color: ti.math.vec3  # type: ignore
    ):
    start = int(ti.floor(center - radius))
    end = int(ti.ceil(center + radius))
    start = ti.math.max(0, start)
    end = ti.math.min(smoke_density_taichi.shape, end)

    for I in ti.grouped(ti.ndrange([start.x, end.x], [start.y, end.y], [start.z, end.z])):
        if (I - center).norm() <= radius:
            smoke_color_taichi[I] = mix(smoke_color_taichi[I], smoke_density_taichi[I], color, density)
            smoke_density_taichi[I] += density

@ti.func
def _draw_point_scalar(
    smoke_density_taichi,
    point,
    density
):
    point_int = int(point)
    point_fraction = point - point_int
    if point_int.x >= 0 and point_int.x < smoke_density_taichi.shape[0] - 1 and point_int.y >= 0 and point_int.y < smoke_density_taichi.shape[1] - 1 and point_int.z >= 0 and point_int.z < smoke_density_taichi.shape[2] - 1:
        strength = density * (1 - point_fraction.x) * (1 - point_fraction.y) * (1 - point_fraction.z)
        if strength > 0:
            I_000 = point_int.x, point_int.y, point_int.z
            smoke_density_taichi[I_000] += strength
        strength = density * (1 - point_fraction.x) * (1 - point_fraction.y) * point_fraction.z
        if strength > 0:
            I_001 = point_int.x, point_int.y, point_int.z + 1
            smoke_density_taichi[I_001] += strength
        strength = density * (1 - point_fraction.x) * point_fraction.y * (1 - point_fraction.z)
        if strength > 0:
            I_010 = point_int.x, point_int.y + 1, point_int.z
            smoke_density_taichi[I_010] += strength
        strength = density * (1 - point_fraction.x) * point_fraction.y * point_fraction.z
        if strength > 0:
            I_011 = point_int.x, point_int.y + 1, point_int.z + 1
            smoke_density_taichi[I_011] += strength
        strength = density * point_fraction.x * (1 - point_fraction.y) * (1 - point_fraction.z)
        if strength > 0:
            I_100 = point_int.x + 1, point_int.y, point_int.z
            smoke_density_taichi[I_100] += strength
        strength = density * point_fraction.x * (1 - point_fraction.y) * point_fraction.z
        if strength > 0:
            I_101 = point_int.x + 1, point_int.y, point_int.z + 1
            smoke_density_taichi[I_101] += strength
        strength = density * point_fraction.x * point_fraction.y * (1 - point_fraction.z)
        if strength > 0:
            I_110 = point_int.x + 1, point_int.y + 1, point_int.z
            smoke_density_taichi[I_110] += strength
        strength = density * point_fraction.x * point_fraction.y * point_fraction.z
        if strength > 0:
            I_111 = point_int.x + 1, point_int.y + 1, point_int.z + 1
            smoke_density_taichi[I_111] += strength

@ti.func
def _draw_point(
    smoke_density_taichi,
    smoke_color_taichi,
    point,
    density,
    color
):
    point_int = int(point)
    point_fraction = point - point_int
    if point_int.x >= 0 and point_int.x < smoke_density_taichi.shape[0] - 1 and point_int.y >= 0 and point_int.y < smoke_density_taichi.shape[1] - 1 and point_int.z >= 0 and point_int.z < smoke_density_taichi.shape[2] - 1:
        strength = density * (1 - point_fraction.x) * (1 - point_fraction.y) * (1 - point_fraction.z)
        if strength > 0:
            I_000 = point_int.x, point_int.y, point_int.z
            smoke_color_taichi[I_000] = mix(smoke_color_taichi[I_000], smoke_density_taichi[I_000], color, strength)
            smoke_density_taichi[I_000] += strength
        strength = density * (1 - point_fraction.x) * (1 - point_fraction.y) * point_fraction.z
        if strength > 0:
            I_001 = point_int.x, point_int.y, point_int.z + 1
            smoke_color_taichi[I_001] = mix(smoke_color_taichi[I_001], smoke_density_taichi[I_001], color, strength)
            smoke_density_taichi[I_001] += strength
        strength = density * (1 - point_fraction.x) * point_fraction.y * (1 - point_fraction.z)
        if strength > 0:
            I_010 = point_int.x, point_int.y + 1, point_int.z
            smoke_color_taichi[I_010] = mix(smoke_color_taichi[I_010], smoke_density_taichi[I_010], color, strength)
            smoke_density_taichi[I_010] += strength
        strength = density * (1 - point_fraction.x) * point_fraction.y * point_fraction.z
        if strength > 0:
            I_011 = point_int.x, point_int.y + 1, point_int.z + 1
            smoke_color_taichi[I_011] = mix(smoke_color_taichi[I_011], smoke_density_taichi[I_011], color, strength)
            smoke_density_taichi[I_011] += strength
        strength = density * point_fraction.x * (1 - point_fraction.y) * (1 - point_fraction.z)
        if strength > 0:
            I_100 = point_int.x + 1, point_int.y, point_int.z
            smoke_color_taichi[I_100] = mix(smoke_color_taichi[I_100], smoke_density_taichi[I_100], color, strength)
            smoke_density_taichi[I_100] += strength
        strength = density * point_fraction.x * (1 - point_fraction.y) * point_fraction.z
        if strength > 0:
            I_101 = point_int.x + 1, point_int.y, point_int.z + 1
            smoke_color_taichi[I_101] = mix(smoke_color_taichi[I_101], smoke_density_taichi[I_101], color, strength)
            smoke_density_taichi[I_101] += strength
        strength = density * point_fraction.x * point_fraction.y * (1 - point_fraction.z)
        if strength > 0:
            I_110 = point_int.x + 1, point_int.y + 1, point_int.z
            smoke_color_taichi[I_110] = mix(smoke_color_taichi[I_110], smoke_density_taichi[I_110], color, strength)
            smoke_density_taichi[I_110] += strength
        strength = density * point_fraction.x * point_fraction.y * point_fraction.z
        if strength > 0:
            I_111 = point_int.x + 1, point_int.y + 1, point_int.z + 1
            smoke_color_taichi[I_111] = mix(smoke_color_taichi[I_111], smoke_density_taichi[I_111], color, strength)
            smoke_density_taichi[I_111] += strength

@ti.kernel
def _draw_line_simple_kernel(  # Draw single-pixel-wide line (Anti-aliasing)
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    start: ti.math.vec3,  # type: ignore
    end: ti.math.vec3,  # type: ignore
    density: float,
    color: ti.math.vec3,  # type: ignore
    end_point: bool,
    step: float
    ):
    length = (end - start).norm()
    point_num = int(ti.ceil(length / step))
    point_num = max(2, point_num)
    point_strength = density * length / (point_num - 1)
    for i in ti.ndrange(point_num if end_point else point_num - 1):
        t = float(i) / (point_num - 1)
        _draw_point(smoke_density_taichi, smoke_color_taichi, start * (1 - t) + end * t, point_strength, color)

def draw_line_simple(  # Draw single-pixel-wide line (Anti-aliasing)
    smoke_density_taichi,
    smoke_color_taichi,
    start,
    end,
    density,
    color,
    step=0.5
    ):
    _draw_line_simple_kernel(smoke_density_taichi, smoke_color_taichi, start, end, density, color, True, step)

def draw_polyline_simple(  # Draw single-pixel-wide line (Anti-aliasing)
    smoke_density_taichi,
    smoke_color_taichi,
    polyline,
    density,
    color,
    step=0.5
):
    if len(polyline) < 2:
        return
    for i in range(len(polyline) - 2):
        _draw_line_simple_kernel(smoke_density_taichi, smoke_color_taichi, polyline[i], polyline[i + 1], density, color, False, step)
    _draw_line_simple_kernel(smoke_density_taichi, smoke_color_taichi, polyline[-2], polyline[-1], density, color, True, step)

# TODO: fill_cylinder

def draw_line(  # Draw line (Anti-aliasing)
    smoke_density_taichi,
    smoke_color_taichi,
    start,
    end,
    radius,
    density,
    color
):
    assert radius == 1  # TODO
    draw_line_simple(smoke_density_taichi, smoke_color_taichi, start, end, density, color)

# TODO: fill_cone

# TODO: draw_arrow

# TODO: draw_circle

@ti.kernel
def _draw_helix_kernel(
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    start: ti.math.vec3,  # type: ignore
    height: float,
    radius: float,
    rounds: float,
    x: ti.math.vec3,  # type: ignore
    y: ti.math.vec3,  # type: ignore
    z: ti.math.vec3,  # type: ignore
    density: float,
    color: ti.math.vec3,  # type: ignore
    step: float
):
    length = (height ** 2 + (2 * ti.math.pi * radius * rounds) ** 2) ** 0.5
    point_num = int(ti.ceil(length / step))
    point_num = max(2, point_num)
    point_strength = density * length / (point_num - 1)
    for i in ti.ndrange(point_num):
        t = float(i) / (point_num - 1)
        _draw_point(smoke_density_taichi, smoke_color_taichi, start + x * (t * height) + y * (radius * ti.cos(2 * ti.math.pi * rounds * t)) + z * (radius * ti.sin(2 * ti.math.pi * rounds * t)), point_strength, color)

def draw_helix(  # Draw helix (Anti-aliasing)
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    start,
    end,
    radius,
    rounds,
    density,
    color,
    initial_direction=None,
    step=0.5
):
    # start = np.array(start, dtype=float)
    # end = np.array(end, dtype=float)
    start = ti.Vector(start)
    end = ti.Vector(end)
    x, y, z = construct_frame(end - start, initial_direction)
    _draw_helix_kernel(
        smoke_density_taichi,
        smoke_color_taichi,
        start,
        (end - start).norm(),
        radius,
        rounds,
        x,
        y,
        z,
        density,
        color,
        step)


# TODO: draw_cubic_bezier_curve

# TODO: draw_cubic_bezier_surface

# TODO: draw_spline

@ti.kernel
def fill_convex(
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    center: ti.math.vec3,  # type: ignore
    face_vectors: ti.template(),  # type: ignore
    radius_consider: float,
    density: float,
    color: ti.math.vec3  # type: ignore
):
    start = int(ti.floor(center - radius_consider))
    end = int(ti.ceil(center + radius_consider))
    start = ti.math.max(0, start)
    end = ti.math.min(smoke_density_taichi.shape, end)

    for I in ti.grouped(ti.ndrange([start.x, end.x], [start.y, end.y], [start.z, end.z])):
        inside = True
        relative_location = I - center
        for i in range(face_vectors.shape[0]):
            if ti.math.dot(relative_location, face_vectors[i]) > 1:
                inside = False
        if inside:
            smoke_color_taichi[I] = mix(smoke_color_taichi[I], smoke_density_taichi[I], color, density)
            smoke_density_taichi[I] += density

def fill_platonic_solid(
    smoke_density_taichi,
    smoke_color_taichi,
    center,
    radius,
    face_num,
    density,
    color,
    transform=None
):
    if transform is None:
        transform = np.eye(3)

    phi = (1 + np.sqrt(5)) / 2
    face_vectors = {
        4: [
            [1, 1, 1],
            [-1, -1, 1],
            [-1, 1, -1],
            [1, -1, -1]
        ],
        6:[
            [1, 0, 0], [-1, 0, 0],
            [0, 1, 0], [0, -1, 0],
            [0, 0, 1], [0, 0, -1]
        ],
        8:[
            [1, 1, 1],
            [1, 1, -1],
            [1, -1, 1],
            [1, -1, -1],
            [-1, 1, 1],
            [-1, 1, -1],
            [-1, -1, 1],
            [-1, -1, -1]
        ],
        12:[
            [0, 1, phi], [0, -1, phi], [0, 1, -phi], [0, -1, -phi],
            [1, phi, 0], [-1, phi, 0], [1, -phi, 0], [-1, -phi, 0],
            [phi, 0, 1], [-phi, 0, 1], [phi, 0, -1], [-phi, 0, -1]
        ],
        20:[
            [1, 1, 1], [1, 1, -1], [1, -1, 1], [1, -1, -1],
            [-1, 1, 1], [-1, 1, -1], [-1, -1, 1], [-1, -1, -1],
            [0, 1 / phi, phi], [0, 1 / phi, -phi], [0, -1 / phi, phi], [0, -1 / phi, -phi],
            [phi, 0, 1 / phi], [phi, 0, -1 / phi], [-phi, 0, 1 / phi], [-phi, 0, -1 / phi],
            [1 / phi, phi, 0], [1 / phi, -phi, 0], [-1 / phi, phi, 0], [-1 / phi, -phi, 0]
        ]
    }[face_num]
    face_vectors = np.array(face_vectors, dtype=float)
    face_vectors /= np.sum(face_vectors ** 2, axis=-1)[:, np.newaxis] ** 0.5  # Normalize
    face_vectors /= radius
    face_vectors @= np.linalg.inv(transform)
    face_vectors_taichi = ti.Vector.field(3, ti.f32, shape=[len(face_vectors)])
    face_vectors_taichi.from_numpy(face_vectors)

    fill_convex(
        smoke_density_taichi,
        smoke_color_taichi,
        center,
        face_vectors_taichi,
        radius * np.max(np.linalg.eigvals(transform)) * {  # Vertex-to-center to face-center-to-center distance ratio in Platonic solids
            4: 3,
            6: 3 ** 0.5,
            8: 3 ** 0.5,
            12: 3 ** 0.5 / phi * ((5 - 5 ** 0.5) / 2) ** 0.5,
            20: 3 ** 0.5 / phi
        }[face_num],
        density,
        color)

# TODO: Plot function

@ti.kernel
def _draw_identical_particles_scalar_kernel(
    field_taichi: ti.template(),  # type: ignore
    particles_taichi: ti.template(),  # type: ignore
    density: float
):
    for point in particles_taichi:
        _draw_point_scalar(field_taichi, point, density)

@ti.kernel
def _draw_particles_scalar_kernel(
    field_taichi: ti.template(),  # type: ignore
    particles_taichi: ti.template(),  # type: ignore
    densities_taichi: ti.template(),  # type: ignore
):
    for i in ti.ndrange(particles_taichi.shape[0]):
        _draw_point_scalar(field_taichi, particles_taichi[i], densities_taichi[i])

@ti.kernel
def _draw_identical_particles_kernel(  # Draw particles with single color and density (Anti-aliasing)
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    particles_taichi: ti.template(),  # type: ignore
    density: float,
    color: ti.math.vec3  # type: ignore
):
    for point in particles_taichi:
        _draw_point(smoke_density_taichi, smoke_color_taichi, point, density, color)

@ti.kernel
def _draw_particles_kernel(  # Draw particles (Anti-aliasing)
    smoke_density_taichi: ti.template(),  # type: ignore
    smoke_color_taichi: ti.template(),  # type: ignore
    particles_taichi: ti.template(),  # type: ignore
    densities_taichi: ti.template(),  # type: ignore
    colors_taichi: ti.template(),  # type: ignore
):
    for i in ti.ndrange(particles_taichi.shape[0]):
        _draw_point(smoke_density_taichi, smoke_color_taichi, particles_taichi[i], densities_taichi[i], colors_taichi[i])

def is_vector(a):
    if isinstance(a, list) and len(a) == 3:  # List
        return True
    if isinstance(a, ti.Vector):  # Taichi vector
        return True
    if isinstance(a, ti.Field):  # Taichi fiels or taichi vector field
        return False
    if hasattr(a, '__len__'):
        if len(a) != 3:
            return False
        if hasattr(a, 'shape'):
            if len(a.shape) == 1:  # Array with shape [3]
                return True
    return False

def draw_particles(  # Draw particles (Anti-aliasing)
    smoke_density_taichi,
    particles,
    smoke_color_taichi=None,
    densities=1.,
    colors=[1, 1, 1],
):
    if not isinstance(particles, ti.Field):
        particles_numpy = particles
        particles = ti.Vector.field(3, dtype=ti.f32, shape=particles.shape[:-1])
        particles.from_numpy(particles_numpy)
    if not type(densities) in [int, float]:
        if not isinstance(densities, ti.Field):  # Numpy array
            densities_numpy = densities
            densities = ti.field(dtype=ti.f32, shape=densities.shape)
            densities.from_numpy(densities_numpy)
    if not is_vector(colors):
        if not isinstance(colors, ti.Field):  # Numpy array
            colors_numpy = colors
            colors = ti.Vector.field(3, dtype=ti.f32, shape=colors.shape[:-1])
            colors.from_numpy(colors_numpy)
    if not smoke_color_taichi is None:
        if isinstance(densities, ti.Field) ^ isinstance(colors, ti.Field):
            if isinstance(densities, ti.Field):  # colors is a value. Need to convert to a constant taichi field.
                colors_numpy = np.zeros([particles.shape[0], 3])
                for i in range(3):
                    colors_numpy[:, i] = colors[i]
                colors = ti.Vector.field(3, dtype=ti.f32, shape=colors_numpy.shape[:-1])
                colors.from_numpy(colors_numpy)
            else:  # densities is a value. Need to convert to a constant taichi field.
                densities_numpy = np.zeros(particles.shape[0])
                densities_numpy.fill(densities)
                densities = ti.field(dtype=ti.f32, shape=densities_numpy.shape)
                densities.from_numpy(densities_numpy)

    if smoke_color_taichi is None:
        if type(densities) in [int, float]:
            _draw_identical_particles_scalar_kernel(smoke_density_taichi, particles, densities)
        else:
            _draw_particles_scalar_kernel(smoke_density_taichi, particles, densities)
    else:
        if type(densities) in [int, float]:
            _draw_identical_particles_kernel(smoke_density_taichi, smoke_color_taichi, particles, densities, colors)
        else:
            _draw_particles_kernel(smoke_density_taichi, smoke_color_taichi, particles, densities, colors)

# TODO: def gaussian_blur(smoke_density_taichi, smoke_color_taichi, radius)

# TODO: def draw_volume(smoke_density_taichi, smoke_color_taichi, smoke_density_taichi_to_draw, smoke_color_taichi_to_draw)

@ti.kernel
def _gaussian_splatting_kernel(
    smoke: ti.template(),  # type: ignore
    smoke_color: ti.template(),  # type: ignore
    positions: ti.template(),  # type: ignore
    rotations: ti.template(),  # type: ignore
    scales: ti.template(),  # type: ignore
    colors: ti.template(),  # type: ignore
    opacities: ti.template()  # type: ignore
):    
    # Render each particle
    for i in ti.ndrange(positions.shape[0]):
        pos = positions[i]
        scale = scales[i]
        
        # Calculate inversed covariance matrix
        inv_cov3d = compute_covariance_inv(rotations[i], scale)
        
        radius = ti.max(scale.x, scale.y, scale.z) * 3.0
        min_pos = ti.max(ti.cast(ti.floor(pos - radius), ti.i32), 0)
        max_pos = ti.min(ti.cast(ti.ceil(pos + radius), ti.i32), ti.Vector(smoke.shape) - 1)
        if max_pos.x < min_pos.x or max_pos.y < min_pos.y or max_pos.z < min_pos.z:
            continue
        
        for I in ti.grouped(ti.ndrange(
            (min_pos.x, max_pos.x + 1),
            (min_pos.y, max_pos.y + 1),
            (min_pos.z, max_pos.z + 1)
        )):
            delta = float(I) - pos
            density = opacities[i] * ti.exp(-0.5 * delta.dot(inv_cov3d @ delta))
            
            # Mix
            _draw_point(smoke, smoke_color, I, density, colors[i])

def gaussian_splatting(  # Render Gaussian splatting data (Tested with data exported from Jawset Postshot)
    smoke,
    smoke_color,
    data,
    offset=[0, 0, 0],
    rotation_matrix=None,
    rotation_quaternion=None,
    y_upward=False,
    scaling=1
):
    if not rotation_matrix is None:
        rotation_matrix = np.array(rotation_matrix, dtype=float)
        rotation_quaternion = rotation_matrix_to_quaternion(rotation_matrix)
    elif not rotation_quaternion is None:
        rotation_matrix = rotation_quaternion_to_matrix(rotation_quaternion)
    else:
        if y_upward:
            rotation_matrix = np.array([
                [1,  0, 0],
                [0,  0, 1],
                [0, -1, 0]
            ], dtype=float)
        else:
            rotation_matrix = np.eye(3)
        rotation_quaternion = rotation_matrix_to_quaternion(rotation_matrix)

    particle_num = len(data['positions'])
    positions = ti.Vector.field(3, ti.f32, shape=[particle_num])
    rotations = ti.Vector.field(4, ti.f32, shape=[particle_num])
    scales = ti.Vector.field(3, ti.f32, shape=[particle_num])
    colors = ti.Vector.field(3, ti.f32, shape=[particle_num])
    opacities = ti.field(ti.f32, shape=[particle_num])

    positions.from_numpy((data['positions'] @ rotation_matrix.T * scaling + offset))
    rotations.from_numpy(quaternion_multiply(rotation_quaternion, data['rotations'].T).T)
    scales.from_numpy(data['scales'] * scaling)
    colors.from_numpy(data['sh_coeffs'][:, 0])
    opacities.from_numpy(data['opacities'])

    _gaussian_splatting_kernel(
        smoke,
        smoke_color,
        positions,
        rotations,
        scales,
        colors,
        opacities)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="taichi-volume-renderer",
    version="1.6.0",
    author="Shengzhi Wu",
    author_email="e1124755@u.nus.edu",
    description="A python package for real-time GPU volume rendering based on taichi",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ShengzhiWu/taichi-volume-renderer",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'taichi'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Visualization"
    ],
    keywords="taichi volume rendering 3d visualization gaussian splatting point cloud graphics"
)